'-------------------------------------------------------------------------------
'
'' Module: SoftSerial.bas
'
' Copyright 2009 Elba Corp.
' All rights reserved.
'
' Non-exclusive permission to use, copy, modify, and distribute this
' software for any purpose and without fee is hereby granted, provided:
'   1) that the above copyright notice is retained in the source code and,
'   2) that it is understood and agreed that no warranty of any kind,
'      express or implied is provided.
' Use of this code in any manner shall be prima facie evidence of
' acceptance of these terms.
'
' You may contact the author by email at dkinzer@zbasic.net.
'
'-------------------------------------------------------------------------------
'
' Overview:
'
' This module implements "bit-banged" serial I/O, similar in concept
' to the SERIN and SEROUT commands of PBasic.  The advantage of using
' this code over using the ZBasic "software UART" channels is that no
' queues are needed and much higher baud rates are attainable (115.2K 
' baud has been tested, for example).  Additionally, the system isn't
' burdened by the overhead of frequent timer interrupts, a burden that
' is higher with higher baud rates.  The disadvantage is that, since
' there is no input queue, data sent by an external device will be
' missed if it is not being awaited at the time it is sent.  For these
' reasons, this code is recommended for transmit-only applications and
' applications where the remote device only transmits in response to
' a request sent to it.  Another disadvantage is that you lose ability
' to simultaneously transmit and receive as well as to perform other
' tasks during transmission/reception.
'
' For both serial input and serial output, interrupts are disabled for
' the duration of each character.  For this reason, it is advisable to
' use the highest possible baud rate to minimize the impact on the system
' of the interrupt disable period.  In general, interrupts will be disabled
' for approximately 10 bit times.
'
' The code uses carefully calculated delay loops (based on the requested
' baud rate) to implement the bit timing.  Because an UnsignedInteger value
' is used as the loop count for the delay loops, the theoretical minimum
' baud rate is 85 baud (see, however, the caveat in the preceding paragraph).
' The theoretical maximum baud rate is approximately 614K (with a CPU
' frequency of 14.7MHz) but this theoretical maximum has not been tested.
'
' This code does not currently support inverted mode, parity or multiple stop
' bits.  However, it could be modified fairly easily to do so.  If any changes
' are made, the code generated for the bit processing loops must be examined
' carefully and, if necessary, the cycle count values must be changed.
'
'----------------------------------------------------------------------------

' uncomment the line below to include a simple test driver
#define INCLUDE_TEST_DRIVER

'----------------------------------------------------------------------------
'
' This code is carefully crafted to execute a 4-cycle loop a given number
' of times (which must not be zero).  This delay loop is used in the routines
' below to realize the bit timing.
'
Declare Sub delayLoop(ByVal loopCnt as UnsignedInteger)
#c
inline void delayLoop(uint16_t loopCnt)
{
	__asm__ __volatile__
	(
		"1:"			"\n\t"
		"subi %A0, 1"	"\n\t"
		"sbci %B0, 0"	"\n\t"
		"brne 1b"		"\n\t"
		: "=&d" (loopCnt)
		: "0" (loopCnt)
	);
}
#endc

' This constant can be used to calculate the loop count needed to achieve
' a delay of a certain number of CPU cycles.
Private Const DelayLoopCycles as UnsignedLong = 4

'----------------------------------------------------------------------------
'
'' PutSerialByte
'
' Send a byte out a pin at a given baud rate.  Note that the specified pin
' should have been set to the idle state (logic 1) prior to this call.
'
Sub PutSerialByte(ByVal pin as Byte, ByVal c as Byte, ByVal baud as UnsignedLong)
	' compute the PORTx address and masks corresponding to the pin
	Dim addr as UnsignedInteger
	addr = Register.Port(pin).DataAddress
	Dim Volatile port as Byte Based addr
	Dim mask as Byte, notMask as Byte
	mask = PortMask(pin)
	notMask = Not mask

	' for inverted mode, switch the two array values
	Dim bitVal(0 to 1) as Byte
	bitVal(0) = 0
	bitVal(1) = mask

	' check for an invalid pin
	If (mask = 0) Then
		Exit Sub
	End If

	' Assemble the bits to be sent (start bit, 8 data bits, stop bit)
	' This needs to change for inverted mode.
	Dim serData as UnsignedInteger
	serData = Shl(CUInt(c) Or &H100, 1)

	' Compute the loop count delay in the sending loop below to realize
	' the bit time. The first part of the loop plus the loop overhead takes
	' 22 cycles so the remaining delay is the bit time (in cycles) minus
	' that value.
	Const LoopOverheadCycles as UnsignedLong = 22
	Dim bitTime as UnsignedLong		' in CPU cycles
	bitTime = Register.CPUFrequency \ baud
	Dim delayLoopCnt as UnsignedInteger
	delayLoopCnt = CUInt(((bitTime - LoopOverheadCycles) + (DelayLoopCycles \ 2)) \ DelayLoopCycles)

	' Send the bits LSB first.  This loop is written carefully so that
	' the first part of it executes in constant time.  The second part
	' then consumes the remainder of the bit window given the
	' specified baud rate.
	Atomic
		Dim bitCnt as Byte = 10
		Do
			' read the port, clear the bit and set the new value
			port = (port And notMask) Or bitVal(LoByte(serData) And 1)
	
			' prepare for sending the next bit
			serData = Shr(serData, 1)
	
			' delay for the remainder of a bit time
			Call delayLoop(delayLoopCnt)

			' indicate another bit has been transmitted
			bitCnt = bitCnt - 1
		Loop While (bitCnt > 0)
	End Atomic
End Sub

'----------------------------------------------------------------------------
'
'' PutSerialStr
'
' Send the characters of a string out a pin at a given baud rate.  Note that the
' specified pin should have been set to the idle state prior to this call.
'
Sub PutSerialStr(ByVal pin as Byte, ByVal s as String, ByVal baud as UnsignedLong)
	Dim sLen as Integer = Len(s)
	Dim i as Integer
	For i = 1 to sLen
		Call PutSerialByte(pin, Asc(s, i), baud)
	Next i
End Sub

'----------------------------------------------------------------------------
'
'' GetSerialByte
'
' Read a single character from the specified input pin using the given baud
' rate.  The code waits for a start bit and then composes the next eight bits
' into a byte that is returned.
'
Function GetSerialByte(ByVal pin as Byte, ByVal baud as UnsignedLong) as Byte
	GetSerialByte = 0

	' compute the PINx address and mask corresponding to the pin
	Dim addr as UnsignedInteger
	addr = Register.Pin(pin).DataAddress
	Dim Volatile pinVal as Byte Based addr
	Dim mask as Byte
	mask = PortMask(pin)

	' check for an invalid pin
	If (mask = 0) Then
		Exit Function
	End If

	' Define the bit values corresponding to logic levels.
	' Switch these for inverted mode.
	Dim logic0 as Byte, logic1 as Byte
	logic0 = &H00
	logic1 = &H80

	' Compute the loop count delay in the receivng loop below to realize
	' the bit time. The first part of the loop plus the loop overhead takes
	' 12 or 13 cycles (depending on the input state) so the remaining delay
	' is the bit time (in cycles) minus that value.
	Dim bitTime as UnsignedLong		' in CPU cycles
	bitTime = Register.CPUFrequency \ baud
	Dim delayLoopCnt as UnsignedInteger
	Const LoopOverheadCycles as UnsignedLong = 12
	delayLoopCnt = CUInt(((bitTime - LoopOverheadCycles) + (DelayLoopCycles \ 2)) \ DelayLoopCycles)
	
	' Compute a delay loop count to position the first sample roughly in
	' the middle of the least significant data bit.  Assuming that the
	' leading edge of the start bit was accurately detected, the first
	' sample should be 1.5 bit times after the leading edge.
	Dim sampleDelayLoopCnt as UnsignedInteger
	Const SampleCenterAdjust as UnsignedLong = 10
	sampleDelayLoopCnt = CUInt((((bitTime * 3) \ 2) - SampleCenterAdjust) \ DelayLoopCycles)
	Dim loopCnt as UnsignedInteger

	' wait for the start bit (needs to change for inverted mode)
	Do While CBool(pinVal And mask)
	Loop

	Atomic
		Dim bitCnt as Byte = 8

		' delay to put the sample point in the middle of the bit window (1.5 bit times)
		Call delayLoop(sampleDelayLoopCnt)
	
		' Read the bits LSB first.
		Do
			Dim data as Byte
		
			' sample the input pin
			If CBool(pinVal And mask) Then
				data = logic1
			Else
				data = logic0
			End If
			GetSerialByte = Shr(GetSerialByte, 1) Or data
	
			' delay for the remainder of a bit time
			Call delayLoop(delayLoopCnt)

			' one more bit done
			bitCnt = bitCnt - 1
		Loop While (bitCnt > 0)
	End Atomic

	' wait for the stop bit (needs to change for inverted mode)
	Do While Not CBool(pinVal And mask)
	Loop
End Function

'----------------------------------------------------------------------------

#if defined(INCLUDE_TEST_DRIVER)
Sub Main()
	Const serOutPin as Byte = 19
	Const serInPin as Byte = 18
	Const br as UnsignedLong = 115200

	' set the output pin to the idle state
	Call PutPin(serOutPin, 1)
	
	' output some characters
'	Call PutSerialByte(serOutPin, &H55, br)
	Call PutSerialStr(serOutPin, "Hello, world!" & Chr(&H0d) & Chr(&H0a), br)
	
	' get characters and echo them to Com1
	Do
		Dim bval as Byte = GetSerialByte(serInPin, br)
		Debug.Print Chr(bval);
		Call Sleep(1)
	Loop
End Sub
#endif
